@@ -15,7 +15,15 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
+#include "sockets.h"
+#include "downloads.h"
+#include "uploads.h"
+#include "nodes.h"
+#include "misc.h"
+#include "getline.h"
+
 guint32 local_ip = 0;
+gboolean is_firewalled = TRUE;		/* Assume the worst --RAM, 20/12/2001 */
 
 static GSList *sl_incoming = (GSList *) NULL;	/* Track incoming sockets */
 
@@ -32,8 +40,11 @@ void socket_monitor_incoming(void)
 		g_assert(s->last_update);
 		/* We reuse the `node_connecting_timeout' parameter, need a new one? */
 		if (now - s->last_update > node_connecting_timeout) {
-			g_warning("connection from %s timed out (%d bytes read)\n",
+			g_warning("connection from %s timed out (%d bytes read)",
 					  ip_to_gchar(s->ip), s->pos);
+			if (s->pos > 0)
+				dump_hex(stderr, "Connection Header",
+					s->buffer, MIN(s->pos, 80));
 			socket_destroy(s);
 			goto retry;			/* Don't know the internals of lists, retry */
 		}
@@ -78,6 +89,8 @@ void socket_free(struct gnutella_socket *s)
 	}
 	if (s->gdk_tag)
 		gdk_input_remove(s->gdk_tag);
+	if (s->getline)
+		getline_free(s->getline);
 	if (s->file_desc != -1)
 		close(s->file_desc);
 	g_free(s);
@@ -91,7 +104,9 @@ static void socket_read(gpointer data, gint source, GdkInputCondition cond)
 {
 	gint r;
 	struct gnutella_socket *s = (struct gnutella_socket *) data;
-	gint count;
+	guint count;
+	guint parsed;
+	guchar *first;
 
 	//s->type = 0;
 
@@ -100,11 +115,14 @@ static void socket_read(gpointer data, gint source, GdkInputCondition cond)
 		return;
 	}
 
+	g_assert(s->pos == 0);		/* We read a line, then leave this callback */
+
 	count = sizeof(s->buffer) - s->pos - 1;		/* -1 to allow trailing NUL */
 	if (count <= 0) {
 		g_warning
 			("socket_read(): incoming buffer full, disconnecting from %s",
 			 ip_to_gchar(s->ip));
+		dump_hex(stderr, "Leading Data", s->buffer, MIN(s->pos, 256));
 		socket_destroy(s);
 		return;
 	}
@@ -123,78 +141,64 @@ static void socket_read(gpointer data, gint source, GdkInputCondition cond)
 
 	s->last_update = time((time_t *) 0);
 	s->pos += r;
-	*(s->buffer + s->pos) = '\0';		/* Bound strchr() below --RAM */
 
 	/*
-	 * XXX FIXME: need to read all the headers, not just the firts so
-	 * that we read the Range: field if there is one.
+	 * Get first line.
 	 */
-	if (s->pos >= 4 && strchr(s->buffer, '\n')) {
-		/* We've got at least a line */
 
-		if (!strncmp(s->buffer, gnutella_hello, 17)) {
-			/* This is an incoming control connection */
-			struct gnutella_node *n;
+	switch (getline_read(s->getline, s->buffer, s->pos, &parsed)) {
+	case READ_OVERFLOW:
+		g_warning("socket_read(): first line too long, disconnecting from %s",
+			 ip_to_gchar(s->ip));
+		dump_hex(stderr, "Leading Data", s->buffer, MIN(s->pos, 256));
+		socket_destroy(s);
+		return;
+	case READ_DONE:
+		if (s->pos != parsed)
+			memmove(s->buffer, s->buffer + parsed, s->pos - parsed);
+		s->pos -= parsed;
+		break;
+	case READ_MORE:		/* ok, but needs more data */
+	default:
+		g_assert(parsed == s->pos);
+		s->pos = 0;
+		return;
+	}
 
-			gdk_input_remove(s->gdk_tag);
-			s->gdk_tag = 0;
-			if (s->pos > 22)	/* 22 = 17 + "0.4\n\n" */
-				g_warning
-					("incoming node %s sent extra bytes after HELLO\n",
-					 ip_port_to_gchar(s->ip, s->port));
-			else if (s->pos < 22)
-				g_warning("incoming node %s sent short HELLO: \"%s\"\n",
-						  ip_port_to_gchar(s->ip, s->port), s->buffer);
+	/*
+	 * We come here only when we got the first line of data.
+	 *
+	 * Whatever happens now, we're not going to use the existing read
+	 * callback, and we'll no longer monitor the socket via the `sl_incoming'
+	 * list: if it's a node connection, we'll monitor the node, if it's
+	 * an upload, we'll monitor the uplaod.
+	 */
 
-			s->pos = 0;
-			sl_incoming = g_slist_remove(sl_incoming, s);
-			s->last_update = 0;
+	gdk_input_remove(s->gdk_tag);
+	s->gdk_tag = 0;
+	sl_incoming = g_slist_remove(sl_incoming, s);
+	s->last_update = 0;
 
-			n = node_add(s, s->ip, s->port);
+	/*
+	 * Dispatch request.
+	 */
 
-			if (n)
-				s->gdk_tag =
-					gdk_input_add(s->file_desc,
-								  (GdkInputCondition) GDK_INPUT_READ |
-								  GDK_INPUT_EXCEPTION, node_read,
-								  (gpointer) n);
+	first = getline_str(s->getline);
 
-		} else if ((!strncmp(s->buffer, "GET", 3)) || (!strncmp(s->buffer, "HEAD", 4))) {		/* This is an Upload request in HTTP */
-			struct upload *up;
-
-			gdk_input_remove(s->gdk_tag);
-
-			/* XXX initiate state machine to remain here while we did not
-			 * XXX read the full headers; then we'll call upload_add().
-			 * XXX better move to a dedicated callback waiting for all headers.
-			 * XXX		--RAM
-			 */
-
-			up = upload_add(s);
-			sl_incoming = g_slist_remove(sl_incoming, s);
-			s->last_update = 0;
-
-			if (up != NULL) {
-				s->gdk_tag =
-					gdk_input_add(s->file_desc,
-								  (GdkInputCondition) GDK_INPUT_WRITE |
-								  GDK_INPUT_EXCEPTION, upload_write,
-								  (gpointer) up);
-				s->resource.upload = up;
-
-			}
-
-			else {
-				socket_destroy(s);
-				return;
-			}
-		} else {
-			g_warning("socket_read(): Got an unknown incoming connection, "
-				"dropping it.\n");
-			g_warning("socket_read: first 80 chars: %.80s\n", s->buffer);
-
-			socket_destroy(s);
-		}
+	if (0 == strncmp(first, gnutella_hello, gnutella_hello_length))
+		node_add(s, s->ip, s->port);	/* Incoming control connection */
+	else if (0 == strncmp(first, "GET ", 4))
+		upload_add(s);
+	else if (0 == strncmp(first, "HEAD ", 5))
+		upload_add(s);
+	else if (0 == strncmp(first, "GIV ", 4))
+		download_push_ack(s);
+	else {
+		gint len = getline_length(s->getline);
+		g_warning("socket_read(): Got an unknown incoming connection, "
+			"dropping it.");
+		dump_hex(stderr, "First Line", first, MIN(len, 160));
+		socket_destroy(s);
 	}
 }
 
@@ -213,15 +217,20 @@ void socket_connected(gpointer data, gint source, GdkInputCondition cond)
 			node_remove(s->resource.node, "Connection failed");
 		else if (s->type == GTA_TYPE_DOWNLOAD && s->resource.download)
 			download_fallback_to_push(s->resource.download, FALSE);
+		else if (s->type == GTA_TYPE_UPLOAD && s->resource.upload)
+			upload_remove(s->resource.upload, "Connection failed");
 		else
 			socket_destroy(s);
 		return;
 	}
 
 	if (cond & GDK_INPUT_READ) {
-		if (proxy_connections
-			&& s->direction == GTA_CONNECTION_PROXY_OUTGOING) {
+		if (
+			proxy_connections
+			&& s->direction == GTA_CONNECTION_PROXY_OUTGOING
+		) {
 			gdk_input_remove(s->gdk_tag);
+			s->gdk_tag = 0;
 
 			if (socks_protocol == 4) {
 				if (recv_socks(s) != 0) {
@@ -271,11 +280,11 @@ void socket_connected(gpointer data, gint source, GdkInputCondition cond)
 		gint res, option, size = sizeof(gint);
 
 		gdk_input_remove(s->gdk_tag);
+		s->gdk_tag = 0;
 
 		/* Check wether the socket is really connected */
 
-		res =
-			getsockopt(s->file_desc, SOL_SOCKET, SO_ERROR,
+		res = getsockopt(s->file_desc, SOL_SOCKET, SO_ERROR,
 					   (void *) &option, &size);
 
 		if (res == -1 || option) {
@@ -314,53 +323,40 @@ void socket_connected(gpointer data, gint source, GdkInputCondition cond)
 		s->pos = 0;
 		memset(s->buffer, 0, sizeof(s->buffer));
 
+		g_assert(s->gdk_tag == 0);
+
 		switch (s->type) {
 		case GTA_TYPE_CONTROL:
 			{
-				s->gdk_tag =
-					gdk_input_add(s->file_desc,
-								  GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
-								  node_read_connecting, (gpointer) s);
-				node_init_outgoing(s->resource.node);
-				break;
+				struct gnutella_node *n = s->resource.node;
+
+				g_assert(n->socket == s);
+				node_init_outgoing(n);
 			}
+			break;
 
 		case GTA_TYPE_DOWNLOAD:
 			{
-				s->gdk_tag =
-					gdk_input_add(s->file_desc,
-								  GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
-								  download_read, (gpointer) s);
-				download_send_request(s->resource.download);
-				break;
+				struct download *d = s->resource.download;
+
+				g_assert(d->socket == s);
+				download_send_request(d);
 			}
+			break;
 
 		case GTA_TYPE_UPLOAD:
 			{
-				if (!(s->gdk_tag)) {
-					struct upload *up;
+				struct upload *u = s->resource.upload;
 
-					up = upload_add(s);
-					if (up != NULL) {
-						s->gdk_tag = gdk_input_add(s->file_desc,
-												   (GdkInputCondition)
-												   GDK_INPUT_WRITE |
-												   GDK_INPUT_EXCEPTION,
-												   upload_write,
-												   (gpointer) up);
-						s->resource.upload = up;
-					} else
-						socket_destroy(s);
-				}
-				break;
+				g_assert(u->socket == s);
+				upload_push_conf(u);
 			}
+			break;
 
 		default:
-			{
-				g_warning("socket_connected(): Unknown socket type %d !",
-						  s->type);
-				socket_destroy(s);		/* ? */
-			}
+			g_warning("socket_connected(): Unknown socket type %d !", s->type);
+			socket_destroy(s);		/* ? */
+			break;
 		}
 	}
 }
@@ -406,7 +402,6 @@ static void socket_accept(gpointer data, gint source,
 		break;
 
 	default:
-
 		g_warning("socket_accept(): Unknown listning socket type %d !\n",
 				  s->type);
 		socket_destroy(s);
@@ -427,8 +422,7 @@ static void socket_accept(gpointer data, gint source,
 
 	fcntl(sd, F_SETFL, O_NONBLOCK);	/* Set the file descriptor non blocking */
 
-	t = (struct gnutella_socket *)
-		g_malloc0(sizeof(struct gnutella_socket));
+	t = (struct gnutella_socket *) g_malloc0(sizeof(struct gnutella_socket));
 
 	t->file_desc = sd;
 	t->ip = g_ntohl(addr.sin_addr.s_addr);
@@ -436,57 +430,49 @@ static void socket_accept(gpointer data, gint source,
 	t->direction = GTA_CONNECTION_INCOMING;
 	t->type = s->type;
 	t->local_port = s->local_port;
+	t->getline = getline_make();
 
 	switch (s->type) {
 	case GTA_TYPE_CONTROL:
-		{
-			t->gdk_tag =
-				gdk_input_add(sd, GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
-							  socket_read, t);
-			/*
-			 * Whilst the socket is attached to that callback, it has been
-			 * freshly accepted and we don't know what we're going to do with
-			 * it.	Is it an incoming node connection or an upload request?
-			 * Can't tell until we have read enough bytes.
-			 *
-			 * However, we must guard against a subtle DOS attack whereby
-			 * someone would connect to us and then send only one byte (say),
-			 * then nothing.  The socket would remain connected, without
-			 * being monitored for timeout by the node/upload code.
-			 *
-			 * Insert the socket to the `sl_incoming' list, and have it
-			 * monitored periodically.	We know the socket is on the list
-			 * as soon as it has a non-zero last_update field.
-			 *				--RAM, 07/09/2001
-			 */
+		t->gdk_tag =
+			gdk_input_add(sd, GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
+						  socket_read, t);
+		/*
+		 * Whilst the socket is attached to that callback, it has been
+		 * freshly accepted and we don't know what we're going to do with
+		 * it.	Is it an incoming node connection or an upload request?
+		 * Can't tell until we have read enough bytes.
+		 *
+		 * However, we must guard against a subtle DOS attack whereby
+		 * someone would connect to us and then send only one byte (say),
+		 * then nothing.  The socket would remain connected, without
+		 * being monitored for timeout by the node/upload code.
+		 *
+		 * Insert the socket to the `sl_incoming' list, and have it
+		 * monitored periodically.	We know the socket is on the list
+		 * as soon as it has a non-zero last_update field.
+		 *				--RAM, 07/09/2001
+		 */
 
-			sl_incoming = g_slist_prepend(sl_incoming, t);
-			t->last_update = time((time_t *) 0);
-			break;
-		}
-
-	case GTA_TYPE_DOWNLOAD:
-		{
-			if (dbg > 7) printf("Accepting INCOMING CONNECTION for %s\n",
-				s->resource.download->file_name);
-
-			t->resource.download = s->resource.download;
-
-			t->resource.download->socket = t;
-
-			t->gdk_tag =
-				gdk_input_add(sd, GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
-							  download_read, t);
-
-			socket_free(s);		/* Close the listening socket */
-
-			break;
-		}
+		sl_incoming = g_slist_prepend(sl_incoming, t);
+		t->last_update = time((time_t *) 0);
+		break;
 
 	default:
 		g_assert(0);			/* Can't happen */
 		break;
 	}
+
+	/*
+	 * Since someone connected to us, we're not completely firewalled.
+	 * The Gnutella port has at least been opened.
+	 *		--RAM, 20/12/2001
+	 */
+
+	if (is_firewalled) {
+		is_firewalled = FALSE;
+		if (dbg) printf("Got evidence that we're not fully firewalled\n");
+	}
 }
 
 /*
@@ -504,12 +490,11 @@ struct gnutella_socket *socket_connect(guint32 ip, guint16 port, gint type)
 	sd = socket(AF_INET, SOCK_STREAM, 0);
 
 	if (sd == -1) {
-		g_warning("Unable to create a socket (%s)\n", g_strerror(errno));
+		g_warning("Unable to create a socket (%s)", g_strerror(errno));
 		return NULL;
 	}
 
-	s = (struct gnutella_socket *)
-		g_malloc0(sizeof(struct gnutella_socket));
+	s = (struct gnutella_socket *) g_malloc0(sizeof(struct gnutella_socket));
 
 	s->type = type;
 	s->direction = GTA_CONNECTION_OUTGOING;
@@ -525,8 +510,8 @@ struct gnutella_socket *socket_connect(guint32 ip, guint16 port, gint type)
 	fcntl(sd, F_SETFL, O_NONBLOCK);	/* Set the file descriptor non blocking */
 
 	addr.sin_family = AF_INET;
-	s->ip = addr.sin_addr.s_addr = g_htonl(ip);
-	s->port = addr.sin_port = g_htons(port);
+	addr.sin_addr.s_addr = g_htonl(ip);
+	addr.sin_port = g_htons(port);
 
 	/*
 	 * Now we check if we're forcing a local IP, and make it happen if so.
@@ -543,7 +528,7 @@ struct gnutella_socket *socket_connect(guint32 ip, guint16 port, gint type)
 		 * behind a masquerading firewall --RAM.
 		 */
 		(void) bind(sd, (struct sockaddr *) &lcladdr,
-					sizeof(struct sockaddr_in));
+			sizeof(struct sockaddr_in));
 	}
 
 	if (proxy_connections) {
@@ -551,20 +536,19 @@ struct gnutella_socket *socket_connect(guint32 ip, guint16 port, gint type)
 		lcladdr.sin_port = INADDR_ANY;
 
 		(void) bind(sd, (struct sockaddr *) &lcladdr,
-					sizeof(struct sockaddr_in));
+			sizeof(struct sockaddr_in));
 
-		res =
-			proxy_connect(sd, (struct sockaddr *) &addr,
-						  sizeof(struct sockaddr_in));
+		res = proxy_connect(sd, (struct sockaddr *) &addr,
+			sizeof(struct sockaddr_in));
 
 		s->direction = GTA_CONNECTION_PROXY_OUTGOING;
 	} else
-		res =
-			connect(sd, (struct sockaddr *) &addr,
-					sizeof(struct sockaddr_in));
+		res = connect(sd, (struct sockaddr *) &addr,
+			sizeof(struct sockaddr_in));
 
 	if (res == -1 && errno != EINPROGRESS) {
-		g_warning("Unable to connect (%s)\n", g_strerror(errno));
+		g_warning("Unable to connect to %s: (%s)",
+			ip_port_to_gchar(s->ip, s->port), g_strerror(errno));
 		socket_destroy(s);
 		return NULL;
 	}
@@ -573,20 +557,21 @@ struct gnutella_socket *socket_connect(guint32 ip, guint16 port, gint type)
 
 	/* Always keep our IP current, in case of dynamic address */
 	if (guess_local_ip(sd, &local_ip, &s->local_port) == -1) {
-		g_warning("Unable to guess our IP ! (%s)\n", g_strerror(errno));
+		g_warning("Unable to guess our IP ! (%s)", g_strerror(errno));
 		socket_destroy(s);
 		return NULL;
 	}
 
+	g_assert(s->gdk_tag == 0);
+
 	if (proxy_connections)
-		s->gdk_tag =
-			gdk_input_add(sd, GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
-						  socket_connected, s);
+		s->gdk_tag = gdk_input_add(sd,
+			GDK_INPUT_READ | GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
+			socket_connected, s);
 	else
-		s->gdk_tag =
-			gdk_input_add(sd,
-						  GDK_INPUT_READ | GDK_INPUT_WRITE |
-						  GDK_INPUT_EXCEPTION, socket_connected, s);
+		s->gdk_tag = gdk_input_add(sd,
+			GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
+			socket_connected, s);
 
 	return s;
 }
@@ -603,12 +588,11 @@ struct gnutella_socket *socket_listen(guint32 ip, guint16 port, gint type)
 	sd = socket(AF_INET, SOCK_STREAM, 0);
 
 	if (sd == -1) {
-		g_warning("Unable to create a socket (%s)\n", g_strerror(errno));
+		g_warning("Unable to create a socket (%s)", g_strerror(errno));
 		return NULL;
 	}
 
-	s = (struct gnutella_socket *)
-		g_malloc0(sizeof(struct gnutella_socket));
+	s = (struct gnutella_socket *) g_malloc0(sizeof(struct gnutella_socket));
 
 	s->type = type;
 	s->direction = GTA_CONNECTION_LISTENING;
@@ -629,7 +613,7 @@ struct gnutella_socket *socket_listen(guint32 ip, guint16 port, gint type)
 	/* bind() the socket */
 
 	if (bind(sd, (struct sockaddr *) &addr, l) == -1) {
-		g_warning("Unable to bind() the socket on port %u (%s)\n",
+		g_warning("Unable to bind() the socket on port %u (%s)",
 				  port, g_strerror(errno));
 		socket_destroy(s);
 		return NULL;
@@ -637,9 +621,8 @@ struct gnutella_socket *socket_listen(guint32 ip, guint16 port, gint type)
 
 	/* listen() the socket */
 
-	if (listen(sd, 1) == -1) {
-		g_warning("Unable to listen() the socket (%s)\n",
-				  g_strerror(errno));
+	if (listen(sd, 5) == -1) {
+		g_warning("Unable to listen() the socket (%s)", g_strerror(errno));
 		socket_destroy(s);
 		return NULL;
 	}
@@ -651,7 +634,7 @@ struct gnutella_socket *socket_listen(guint32 ip, guint16 port, gint type)
 
 		if (getsockname(sd, (struct sockaddr *) &addr, &option) == -1) {
 			g_warning("Unable to get the port of the socket: "
-				"getsockname() failed (%s).", g_strerror(errno));
+				"getsockname() failed (%s)", g_strerror(errno));
 			socket_destroy(s);
 			return NULL;
 		}
@@ -671,7 +654,11 @@ struct gnutella_socket *socket_listen(guint32 ip, guint16 port, gint type)
 
 static void show_error(char *fmt, ...)
 {
+	va_list args;
 
+	va_start(args, fmt);
+	vfprintf(stderr, fmt, args);
+	va_end(args);
 }
 
 /*
@@ -757,8 +744,8 @@ int send_socks(struct gnutella_socket *s)
 	/* Create the request */
 	thisreq->version = 4;
 	thisreq->command = 1;
-	thisreq->dstport = s->port;
-	thisreq->dstip = s->ip;
+	thisreq->dstport = htons(s->port);
+	thisreq->dstip = htonl(s->ip);
 
 	/* Copy the username */
 	strcpy(realreq + sizeof(struct sockreq),
@@ -878,7 +865,10 @@ int connect_socksv5(struct gnutella_socket *s)
 			return (rc);
 		}
 
-		if ((unsigned short int) buf[1] == 2)
+		if (
+			(unsigned short int) buf[1] == 2 &&
+			socksv5_user && socksv5_user[0]		/* has provided user info */
+		)
 			s->pos++;
 		else
 			s->pos += 3;
@@ -956,8 +946,8 @@ int connect_socksv5(struct gnutella_socket *s)
 		buf[1] = '\x01';		/* Connect request */
 		buf[2] = '\x00';		/* Reserved		*/
 		buf[3] = '\x01';		/* IP version 4	*/
-		memcpy(&buf[4], &s->ip, 4);
-		memcpy(&buf[8], &s->port, 2);
+		*(uint32_t *)(buf + 4) = htonl(s->ip);
+		*(uint16_t *)(buf + 8) = htons(s->port);
 
 		/* Now send the connection */
 		if ((rc = send(sockid, (void *) buf, 10, 0)) <= 0) {
@@ -976,7 +966,7 @@ int connect_socksv5(struct gnutella_socket *s)
 			rc = ECONNREFUSED;
 			return (rc);
 		}
-		printf("connect_socksv5: Step 5, bytes recv'd %i\n", rc);
+		if (dbg) printf("connect_socksv5: Step 5, bytes recv'd %i\n", rc);
 		if (rc < 10) {
 			show_error("Short reply from SOCKS server\n");
 			rc = ECONNREFUSED;
